using System;
using System.Net;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.ServiceModel.Description;
using System.ServiceModel.Dispatcher;
using System.Xml;
using System.Xml.Serialization;

namespace WcfRestPox
{
	public class RestPoxBehavior : IEndpointBehavior, IOperationBehavior, IClientMessageInspector
	{
		private const string QueryStringPropertyName = "QueryString";

		#region IEndpointBehavior Members

		public void Validate(ServiceEndpoint endpoint)
		{
		}

		public void AddBindingParameters(ServiceEndpoint endpoint, BindingParameterCollection bindingParameters)
		{
			foreach (OperationDescription operation in endpoint.Contract.Operations)
				operation.Behaviors.Add(this);
		}

		public void ApplyDispatchBehavior(ServiceEndpoint endpoint, EndpointDispatcher endpointDispatcher)
		{
		}

		public void ApplyClientBehavior(ServiceEndpoint endpoint, ClientRuntime clientRuntime)
		{
			clientRuntime.MessageInspectors.Add(this);
		}

		#endregion

		#region IOperationBehavior Members

		public void Validate(OperationDescription operationDescription)
		{
		}

		public void ApplyDispatchBehavior(OperationDescription operationDescription, DispatchOperation dispatchOperation)
		{
		}

		public void ApplyClientBehavior(OperationDescription operationDescription, ClientOperation clientOperation)
		{
			clientOperation.Formatter = new RestPoxMessageFormatter(
				operationDescription.Messages[0],
				operationDescription.Messages[1]);
			clientOperation.SerializeRequest = true;
			clientOperation.DeserializeReply = true;
		}

		public void AddBindingParameters(OperationDescription operationDescription,
										 BindingParameterCollection bindingParameters)
		{
		}

		#endregion

		#region IClientMessageInspector Members

		public object BeforeSendRequest(ref Message request, IClientChannel channel)
		{
			HttpRequestMessageProperty httpRequest = new HttpRequestMessageProperty();

			// TODO: this should be configurable for each operation
			httpRequest.Method = WebRequestMethods.Http.Get;

			// Suppress the body (GET requests cannot have body)
			httpRequest.SuppressEntityBody = true;

			// Suppress the Content-Type header
			httpRequest.Headers[HttpRequestHeader.ContentType] = null;

			// Set the query string configured by the message formatter
			httpRequest.QueryString = (string)request.Properties[QueryStringPropertyName];

			// Rewrite the To address
			request.Headers.To = new Uri(channel.RemoteAddress.Uri, request.Headers.To);

			request.Properties.Add(HttpRequestMessageProperty.Name, httpRequest);
			return null;
		}

		public void AfterReceiveReply(ref Message reply, object correlationState)
		{
		}

		#endregion

		#region IClientMessageFormatter Implementation

		private class RestPoxMessageFormatter : IClientMessageFormatter
		{
			private readonly MessageDescription requestInfo;
			private readonly MessageDescription responseInfo;

			public RestPoxMessageFormatter(MessageDescription requestInfo, MessageDescription responseInfo)
			{
				this.requestInfo = requestInfo;
				this.responseInfo = responseInfo;
			}

			public Message SerializeRequest(MessageVersion messageVersion, object[] parameters)
			{
				Message message = Message.CreateMessage(messageVersion, null);

				// This relative address will be replaced later with an absolute one
				message.Headers.To = new Uri(requestInfo.Action, UriKind.Relative);

				if (requestInfo.Body.Parts.Count != parameters.Length)
					throw new InvalidOperationException("Number of parameters received does not match with the message description");

				string[] queryParts = new string[parameters.Length];
				for (int i = 0; i < parameters.Length; i++)
				{
					if (parameters[i] != null)
					{
						string value;
						// TODO: this special formatters should be pluggable
						if (parameters[i] is DateTime)
						{
							value = ((DateTime)parameters[i]).ToUniversalTime().ToString("o");
						}
						else if (parameters[i] is string[])
						{
							value = string.Join(" ", ((string[])parameters[i]));
						}
						else
						{
							value = parameters[i].ToString();
						}

						queryParts[i] = string.Format("{0}={1}",
													  requestInfo.Body.Parts[i].Name,
													  Uri.EscapeDataString(value));
					}
				}

				// This will be later placed in the target address
				message.Properties.Add(QueryStringPropertyName, string.Join("&", queryParts));

				return message;
			}

			public object DeserializeReply(Message message, object[] parameters)
			{
				XmlDictionaryReader reader = message.GetReaderAtBodyContents();
				XmlSerializer serializer = new XmlSerializer(responseInfo.Body.ReturnValue.Type);
				object response = serializer.Deserialize(reader);
				return response;
			}
		}

		#endregion
	}
}